################################## 
# Description: Receives records after subscribing to the MSG 
#      +  Validates the records
#      +  Insert the records on the corresponding db view
#      +  According to Grid Monitoring Probe Specification v0.91
# Author: Piotr Nyczyk(base code), Daniel Rodrigues
# Institute: CERN
# Created: January 08
# License:
# Version: 0.9.2-1, 7 Jul 2008 
# $Id:$ 
################################## 

import os
import sys
import time
import threading
import cx_Oracle
from ConfigParser import ConfigParser,NoOptionError
from Queue import Queue
import stomp
import logging
try :
    from logging.handlers import RotatingFileHandler
except ImportError :
    from logging import RotatingFileHandler




config=None

class StatsCollector:
    def __init__(self):
        self.stats={}
        
    def setFlag(self,stat,flag,state=True):
        if not self.stats.has_key(stat):
            self.stats[stat]={}
        self.stats[stat][flag]=state
            
    def updateCounter(self,stat,increment=1):
        if not self.stats.has_key(stat):
            self.stats[stat]=0
        self.stats[stat]+=increment
    
    def setCounter(self,stat,value=0):
        self.stats[stat]=value

    def dumps(self):
        l=self.stats.keys()
        l.sort()
        s=""
        for k in l:
            s+="%s: %s\n"%(k,str(self.stats[k]))
        return s

stats=StatsCollector()
        
def read_config(configFile):
    global config
    config=ConfigParser()
    if os.path.exists(configFile) :
        config.read(configFile)
    else:
        raise OSError("Could not open config : %s"%configFile)

def log_config() :
    'Changes logger settings after reading configuration file.' 
    debugLevel = int( config.get('logging','debug-level') )
    if debugLevel in (0,10,20,30,40,50) :
        logger.setLevel( debugLevel )

def get_message_class(msgClass):
    try:
        attrs=map(str.lower,config.get('msg-classes',msgClass+'_dbattrs').split())
        table=config.get('msg-classes',msgClass+'_dbtable')
    except NoOptionError:
        logger.error('Unknow message class: %s!\n'%msgClass)
        sys.exit(1)
    return table,attrs

def create_insert_sql(table,attrs):
    colstr=','.join(attrs)
    valstr=','.join(map(lambda x: ':'+x,attrs))
    sql="insert into %s(%s) values(%s)"%(table,colstr,valstr)
    return sql

def validate_record(record,attrs,cursor):
    empty=True
    for attr in attrs:
        if attr not in record or record[attr] is None :
            record[attr]=None
        else:
            if len(record[attr])>1000:
                v=cursor.var(cx_Oracle.LONG_STRING,len(record[attr]))
                v.setvalue(0,record[attr])
                record[attr]=v
            empty=False
    for key in record.keys():
        if key not in attrs:
            del record[key]
    return not empty
            
def read_record(linesQueue):
    logger.debug( 'reading record' )
    record={}
    key=''
    value=''
    multiline=False
    line=linesQueue.get()
    while line and line!='EOT':
        if multiline:
            value+='\n'+line
        elif ':' in line:
            key,value=map(str.strip, line.split(':',1))
            key=key.lower()
            if key=='detailsdata':
                multiline=True
            else:
                record[key]=value
        else:
            logger.error('Syntax error: '+line)
        line=linesQueue.get()
    if multiline:
        record[key]=value
    return record

class DBCommiter(threading.Thread):
    def __init__(self,conn,commitDelay):
        threading.Thread.__init__(self,name='DBCommiter')
        self.setDaemon(True)
        self.conn=conn
        self.lock=threading.Lock()
        self.commitDelay=commitDelay
        self.nRecords={}#0
        self.running=True

    def run(self):
        logger.debug('Running DBCommiter.')
        while self.running:
            time.sleep(self.commitDelay)
            self.acquire()
            self.conn.commit()
            self.release()
            #sys.stderr.write('%s: commited %d records\n'%(time.ctime(),self.nRecords))
            #self.nRecords=0

    def acquire(self):
        stats.setFlag('commiter_lock_waiting',threading.currentThread().getName(),True)
        self.lock.acquire()
        stats.setFlag('commiter_lock_waiting',threading.currentThread().getName(),False)
        stats.setFlag('commiter_lock_acquired',threading.currentThread().getName(),True)

    def release(self):
        self.lock.release()
        stats.setFlag('commiter_lock_acquired',threading.currentThread().getName(),False)
        

def format_record(record):
    multiline=None
    s=''
    for key,value in record.items():
        if not value:
            continue
        if key.lower() == 'detailsdata' :
            multiline='%s: %s\n'%(key,str(value)) 
        else:
            s+='%s: %s\n'%(key,str(value))
    if multiline:
        s+=multiline 
        #s+='%s: EOT\n%s'%multiline
    return s+'EOT\n#'

class BrokerLogHandler(logging.Handler):
    def __init__(self,destination,level=0):
        logging.Handler.__init__(self,level)
        self.destination=destination
    def emit(self,record):
        broker.send(self.destination,stomp.Message(record.getMessage()))


    

class DBInserter(threading.Thread):
    def __init__(self,commiter,msgClass,queueSize=1000):
        threading.Thread.__init__(self,name='DBInserter_'+msgClass)
        self.setDaemon(True)
        self.commiter=commiter
        self.cur=commiter.conn.cursor()
        self.msgClass=msgClass
        self.table,self.attrs=get_message_class(msgClass)
        self.insert_sql=create_insert_sql(self.table,self.attrs)
        self.queue=Queue(queueSize)
        logname='rejected_'+msgClass
        self.rejectedLog=logging.getLogger(logname)
        self.rejectedLog.setLevel(logging.DEBUG)
        logfile=config.get('logging','rejected-location')+'/'+logname+'.log'
 
        # Does'n exist in Python 2.3
        hdlr=RotatingFileHandler(filename=logfile,
                                         maxBytes=int(config.get('logging','rejected-size')),
                                         backupCount=int(config.get('logging','rejected-count')))

        #hdlr=logging.FileHandler(logfile)
        
        form=logging.Formatter('# %(asctime)s %(message)s')
        hdlr.setFormatter(form)
        self.rejectedLog.addHandler(hdlr)
        destination=config.get('monitoring','rejected-destination')+'.'+msgClass
        hdlr=BrokerLogHandler(destination)
        self.rejectedLog.addHandler(hdlr)

    def run(self):
        logger.debug('Running DBInserter')
        while True:
            stats.updateCounter(self.getName()+'_read_record',1)
            #logger.info("reading record")
            record=read_record(self.queue)
            stats.updateCounter(self.getName()+'_read_record',-1)
            try:
                #logger.info("iacquiring commiter.")
                self.commiter.acquire()
                if not validate_record(record,self.attrs,self.cur):
                    self.commiter.release()
                    logger.error("BAD RECORD\n")
                    self.rejectedLog.error('VALIDATION ERROR\n'+format_record(record))
                    continue
                logger.debug('executing insert_Sql \n %s::%s' %(self.insert_sql,str(record)) )
                self.cur.execute(self.insert_sql,record)
                self.commiter.nRecords.setdefault(self.msgClass,0)
                self.commiter.nRecords[self.msgClass]+=1
                self.commiter.release()
                logger.debug('successfuly executed insert query')
            except Exception, e:
                self.commiter.release()
                logger.exception("Got exception: %s While inserting record: %s\n\n"%(str(e),str(record)))
                logger.debug("Insert query and values: %s::%s " %(self.insert_sql,str(record)) ) 
                estr=''.join(['# %s\n'%l for l in str(e).split('\n')])
                self.rejectedLog.error(estr+format_record(record))

    def receive(self,msg):
        logger.debug('message received' + str(msg.header))
        stats.updateCounter(self.getName()+'_receive',1)            
        if msg.header.has_key('msgencodedtime') :
            self.queue.put('msgencodedtime:' + msg.header['msgencodedtime'])
        if msg.header.has_key('msgdecodedtime') :
            self.queue.put('msgdecodedtime:' + msg.header['msgdecodedtime'])
        for line in msg.body.strip().split('\n'):
            self.queue.put(line)
        stats.updateCounter(self.getName()+'_receive',-1)            

commiter=None
inserters=[]
broker=None

def start():
    global commiter
    global inserters
    global broker
    
    db=cx_Oracle.connect(config.get('db','database'))
    
    broker_host=config.get('stomp','broker_host')
    broker_port=int(config.get('stomp','broker_port'))
    client_id=config.get('stomp','client_id')
    logger.info("Starting connection")
    broker=stomp.Connection(broker_host,broker_port,client_id=client_id)

    logger.info("Starting Commiter: broker connection: " + str(broker))
    commiter=DBCommiter(db,1)
    commiter.start()

    logger.info("Starting Inserters")
    inserters={}
    for msgClass in config.get('msg-classes','classes').split():
        msgDest=config.get('msg-classes',msgClass+'_destination')
        msgInserter=DBInserter(commiter,msgClass)
        msgInserter.start()
        broker.subscribe(msgInserter.receive,msgDest,subscriptionName=msgClass)
        inserters[msgClass]=msgInserter
        logger.info("Created DBInserter for message class %s using destination %s"%(msgClass,msgDest))


class MsgLogger :
    'Provides the necessary tools to log back into the messaging system'

    def __init__(self, enabled) :
        self.enabled = True
        if not( enabled == '1' or enabled == 'True' ) :
           self.enabled = False
           return
        self.broker_host=config.get('stomp','broker_host')
        self.broker_port=int(config.get('stomp','broker_port'))
        self.client_id=config.get('stomp','client_id')
        self.monitoringTopic=config.get('monitoring','destination')
        self.monitoringTime=time.ctime()

    def enable(self ) :
        self.enabled=True
    
    def log( self,keyValueMap ) :
        'Gets a map of key:value pairs to be sent'
        if not self.enabled :
            return
      
        monitoringInformation="consumerId:" + self.client_id + "\n"
        monitoringInformation+="lastMonitoringTime:" + self.monitoringTime + "\n"
        self.monitoringTime=time.ctime()
        monitoringInformation+="monitoringTime:" + self.monitoringTime + "\n"
        for key in keyValueMap.keys() :
            monitoringInformation+="topic_" + str(key) + ":" + str(keyValueMap[key]) + "\n"
        broker.send(self.monitoringTopic, stomp.Message(monitoringInformation))
       
    def error( self, message ) :
        'Sends back a message with the error description'
        if not self.enabled:
            return
        monitoringInformation="consumerId:" + self.client_id + "\n"
        monitoringInformation+="errorTime:" + time.ctime() + "\n"
        monitoringInformation+="detailsData:" + message + "\n"
        broker.send(self.monitoringTopic, stomp.Message(monitoringInformation) )



def stats_loop(refresh=5):
    'Infinite loop for logging statistics to default logger, and if enabled, sending to a configured logging location.'

    logger.debug('Running stats loop. Refreshing data every %d sec.' %refresh)

    msgLogger = MsgLogger( config.get('monitoring','enabled') )
    monRefresh=int(config.get('monitoring','refresh'))

    monitoringLastRefreshTime=time.time()
    msgMonInfo={}
    
    while True:
        if commiter.nRecords:
            for msgClass in commiter.nRecords.keys() :
                if commiter.nRecords[msgClass] :
                    logger.info("Got %s records from %s" % (str(commiter.nRecords[msgClass]),msgClass))
                    msgMonInfo.setdefault(msgClass,0)
                    msgMonInfo[msgClass]+=commiter.nRecords[msgClass]
                    logger.debug('msgMonInfor[msgClass] %s' % str(msgMonInfo[msgClass]))
                commiter.nRecords[msgClass]=0
       
        if (time.time() > (monRefresh + monitoringLastRefreshTime)) :
           currentTime = time.time()
           flagNoRecords = True
           for msgClass in msgMonInfo.keys() :
               if msgMonInfo[msgClass] > 0 :
                    flagNoRecords = False
           if not flagNoRecords :
               msgLogger.log(msgMonInfo);    

           flagNoRecords=True;
           monitoringLastRefreshTime = currentTime
           msgMonInfo={}
 
        time.sleep(refresh)

def setup_logging(logfile, runAsDaemon, verbose):
    global logger

    logger=logging.getLogger('main')
    if verbose:
        logger.setLevel(logging.DEBUG)
    else :
        logger.setLevel(logging.INFO)

    fileHdlr=RotatingFileHandler(filename=logfile,
                                     maxBytes=int(config.get('logging','log-size')),
                                     backupCount=int(config.get('logging','log-count')))
    #fileHdlr=logging.FileHandler(logfile)
 
    format=logging.Formatter('%(asctime)s - %(name)s - %(levelname)s %(message)s ')
    fileHdlr.setFormatter(format)
    logger.addHandler(fileHdlr)
    if not runAsDaemon: # Also log to STDOUT
        stdoutHdlr=logging.StreamHandler()
        stdoutHdlr.setFormatter(format)
        logger.addHandler(stdoutHdlr)


def daemonize(pidFile):
    # do the UNIX double-fork magic, see Stevens' "Advanced 
    # Programming in the UNIX Environment" for details (ISBN 0201563177)
    try: 
        pid = os.fork() 
        if pid > 0:
            # exit first parent
            sys.exit(0) 
    except OSError, e: 
        print >>sys.stderr, "fork #1 failed: %d (%s)" % (e.errno, e.strerror) 
        sys.exit(1)

   # decouple from parent environment
    os.chdir("/") 
    os.setsid() 
    os.umask(0) 

    # do second fork
    try: 
        pid = os.fork() 
        if pid > 0:
            # write pid file
            PID = open(pidFile,"w")
            print >>PID,pid
            PID.close()
            # exit from second parent, print eventual PID before
            sys.exit(0) 
    except OSError, e: 
        print >>sys.stderr, "fork #2 failed: %d (%s)" % (e.errno, e.strerror) 
        sys.exit(1) 
    
    #close stdin/stdout/stderr
    map(os.close,range(3))

    # write pid file
    PID = open(pidFile,"w")
    print >>PID,pid
    PID.close()

def main(configFile, logFile, pidFile, runAsDaemon=True, verbose=False):
    if runAsDaemon:
        daemonize(pidFile)
    read_config(configFile)
    setup_logging(logFile, runAsDaemon, verbose)
    logger.info("Logging setup from " + str(configFile) + " into " + str(logFile) )

    rejected_location = config.get('logging','rejected-location')
    if rejected_location != None:
        try: 
            os.makedirs(rejected_location);
        except os.error:
            pass

    logger.debug("Logging configuration")
    log_config()
    logger.debug("Logging configured")
    start()
    stats_loop()

def consume_rejected():
    read_config()
    setup_logging(daemon=False)
    logger.error('consume_rejected function is not implemented yet!')
    pass

    
            
        
        
        
