################################## 
# Description: Implements STOMP protocol 
#      +  implements Connection
#      +  Unmarshalls message
# Author: Piotr Nyczyk(base code), Daniel Rodrigues
# Institute: CERN
# Created: January 08
# License: 
# $Id:$ 
################################## 

import sys
import uuid
import socket
import threading
import time
import logging

#logging.basicConfig()
#logging.root.setLevel(logging.DEBUG)
logger=logging.getLogger('main.stomp')

class ConnectionError(Exception):
    pass

class ProtocolError(Exception):
    pass

class Message:
    ''' Message instance ''' 
    def __init__(self,body,header={}):
        self.body=body
        self.header=dict(header)

    def __str__(self):
        return self.body

    def __repr__(self):
        if self.header:
            return 'Message'+repr((self.body,self.header))
        else:
            return 'Message'+repr((self.body,))

class ProtocolMessage:
    ''' STOMP protocol message implementation in order to 
           dump: send commands across the wire e.g., {CONNECT, SUBSCRIBE,... }. 
           load: receive and unmarshall incoming messages.
    '''
    def __init__(self,cmd,header={},body=''):
        self.cmd=cmd
        self.header=header
        self.body=body

    def __repr__(self):
        return 'ProtocolMessage'+repr((self.cmd,self.header,self.body))
    
    def dump(self,outfile):
        outfile.write(self.cmd+'\n')
        outfile.write(''.join(map(lambda x: '%s:%s\n'%x,self.header.items())))
        outfile.write('\n')
        if self.body:
            outfile.write(self.body)
        outfile.write('\x00\n')
        outfile.flush()

    def load(infile):
        # read command
        cmd=infile.readline().strip()
        #read header
        header={}
        #header={'msgdecodedtime':repr(time.time())}
        while True:
            l=infile.readline().strip()
            if l:
                k,v=l.split(':',1)
                header[k.strip()]=v.strip()
            else:
                break
        # read body
        body=''
        if header.has_key('content-length'):
            body=infile.read(int(header['content-length']))
            #read null byte line
            if infile.readline().strip()!='\x00':
                raise ProtocolError
        else:
            while True:
                data=infile.readline()
                if not data:
                    break
                sdata=data.strip()
                if sdata and sdata[-1]=='\x00':
                    body+=data.strip()[:-1]
                    break
                body+=data
        return ProtocolMessage(cmd,header,body)
    load=staticmethod(load)
        
class Connection:
    ''' Connects to the Broker and implements receive, subscribe, send.
        1 - CONNECT
        2 - if Connected, starts receiverThread
        3 - 
    '''  
    class ReceiverThread(threading.Thread):
        def __init__(self,connection):
            threading.Thread.__init__(self)
            self.setDaemon(True)
            self.connection=connection
            self.running=True

        def run(self):
            logger.debug( 'Running ReceiverThread [' + str(self) + ']' )
            while self.running:
                self.connection.receive(ProtocolMessage.load(self.connection.broker))
            logger.debug( 'Stopped ReceiverThread [' + str(self) + ']' )

    
    def __init__(self,host,port,client_id=None):
        logger.debug('Initiating Connection on ' + str(host) + ':' + str(port) + ':' + str(client_id) )
        s=socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.connect((host,port))
        self.broker=s.makefile()
        header={}
        if client_id:
            header['client-id']=client_id
        ProtocolMessage('CONNECT',header).dump(self.broker)
        m=ProtocolMessage.load(self.broker)
        if m.cmd!='CONNECTED':
            raise ConnectionError
        self.subsByID={}
        self.subsByDest={}
        self.receiver=Connection.ReceiverThread(self)
        self.receiver.start()
        self.connected=True

    def __del__(self):
        self.disconnect()
        
    def receive(self,msg):
        #goout=open('/var/log/msg-consume2oracle.log.s','a')
        #sys.stdout.write('Received: message: ' + msg.cmd)
        #goout.write('Received: message: ' + str(msg.cmd) + ' ,\n')
        #goout.write('len:' + str(len(msg.body)) +'\n' )
        #goout.close();
        if msg.cmd=='MESSAGE':
            id=msg.header['subscription']
            m=Message(msg.body,msg.header)
            for subs in self.subsByID[id]:
                try:
                    subs(m)
                except Exception,e:
                    sys.stderr.write("Exception in receiver: "+str(sys.exc_info()))

    def subscribe(self,receiver,destination,selector='',subscriptionName=None):
        logger.debug('Subscribing to ' + str( subscriptionName ) )
        d=(destination,selector)
        if self.subsByDest.has_key(d):
            id=self.subsByDest[d]
        else:
            id=str(uuid.uuid1())
            self.subsByDest[d]=id
        if self.subsByID.has_key(id):
            self.subsByID[id].append(receiver)
        else:
            self.subsByID[id]=[receiver]
            header={'destination':destination,'id':id}
            if selector:
                header['selector']=selector
            if subscriptionName:
                # the spelling error below is IMPORTANT - it's because of ActiveMQ implementation
                header['activemq.subcriptionName']=subscriptionName
            ProtocolMessage('SUBSCRIBE',header).dump(self.broker)

    def unsubscribe(self,destination,selector=''):
        d=(destination,selector)
        if self.subsByDest.has_key(d):
            id=self.subsByDest[d]
        else:
            return
        ProtocolMessage('UNSUBSCRIBE',{'id':id}).dump(self.broker)
        del self.subsByDest[d]
        del self.subsByID[id]
        
        
    def send(self,destination,msg):
        h=dict(msg.header)
        h['destination']=destination
        ProtocolMessage('SEND',h,msg.body).dump(self.broker)

    def disconnect(self):
        if self.connected:
            self.receiver.running=False
            ProtocolMessage('DISCONNECT').dump(self.broker)
            self.connected=False
        
        
